"
 COPYRIGHT (c) 1988 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

"{ Package: 'stx:libbasic' }"

Magnitude subclass:#Character
	instanceVariableNames:'asciivalue'
	classVariableNames:'CharacterTable'
	poolDictionaries:''
	category:'Magnitude-General'
!

!Character class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1988 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

documentation
"
    This class represents characters.

    Notice, that actual character objects are not used when characters
    are stored in strings, symbols etc; 
    these only store a character's asciiValue/codePoint for a more compact representation.
    The word 'asciiValue' is a historic leftover - actually, any integer
    code is allowed and actually used (i.e. characters are not limited to 8bit).

    Single byte Characters are unique; i.e. for every asciiValue (0..255) there exists exactly 
    one instance of Character, which is shared 
    (Character value:xxx checks for this, and returns a reference to an existing instance).
    Other characters (i.e. asciivalue > 255) are not guaranteed to be shared; 
    i.e. these might be created as required (actually, do NOT depend on which characters are and
    which are not shared; always compare using #= if there is any chance of a non-ascii character
    being involved.
    
    This means: you may compare characters using #== ONLY IFF you are certain,
    that the characters ranges is 0..255. 
    Otherwise, you HAVE TO compare using #=. (if in doubt, always compare using #=).
    Sorry for this inconvenience, but it is (practically) impossible to keep
    the possible maximum of 2^32 characters (Unicode) around, for that convenience alone.


    Methods marked as (JS) come from the manchester Character goody
    (CharacterComparing) by Jan Steinman, which allow Characters to be used as
    Interval elements (i.e. ($a to:$z) do:[...] ); 
    They are not a big deal, but convenient add-ons.
    Some of these have been modified a bit.

    WARNING: characters are known by compiler and runtime system -
             do not change the instance layout. 

    Also, although you can create subclasses of Character, the compiler always
    creates instances of Character for literals ...
    ... and other classes are hard-wired to always return instances of characters
    in some cases (i.e. String>>at:, Symbol>>at: etc.).
    Therefore, it may not make sense to create a character-subclass.

    Case Mapping in Unicode:
        There are a number of complications to case mappings that occur once the repertoire 
        of characters is expanded beyond ASCII.

        * Because of the inclusion of certain composite characters for compatibility, 
          such as U+01F1 'DZ' capital dz, there is a third case, called titlecase, 
          which is used where the first letter of a word is to be capitalized 
          (e.g. Titlecase, vs. UPPERCASE, or lowercase).
          For example, the title case of the example character is U+01F2 'Dz' capital d with small z.

        * Case mappings may produce strings of different length than the original.
          For example, the German character U+00DF small letter sharp s expands when uppercased to 
          the sequence of two characters 'SS'. 
          This also occurs where there is no precomposed character corresponding to a case mapping.
          *** This is not yet implemented (in 5.2) ***

        * Characters may also have different case mappings, depending on the context.
          For example, U+03A3 capital sigma lowercases to U+03C3 small sigma if it is not followed 
          by another letter, but lowercases to 03C2 small final sigma if it is.
          *** This is not yet implemented (in 5.2) ***

        * Characters may have case mappings that depend on the locale.
          For example, in Turkish the letter 0049 'I' capital letter i lowercases to 0131 small dotless i.
          *** This is not yet implemented (in 5.2) ***

        * Case mappings are not, in general, reversible.
          For example, once the string 'McGowan' has been uppercased, lowercased or titlecased, 
          the original cannot be recovered by applying another uppercase, lowercase, or titlecase operation.

    Collation Sequence:
        *** This is not yet implemented (in 5.2) ***

    [author:]
        Claus Gittinger

    [see also:]
        String TwoByteString Unicode16String Unicode32String
        StringCollection
"
! !

!Character class methodsFor:'instance creation'!

basicNew
    "catch new - Characters cannot be created with new"

    ^ self error:'Characters cannot be created with new'
!

digitValue:anInteger
    "return a character that corresponds to anInteger.
     0-9 map to $0-$9, 10-35 map to $A-$Z"

    |val "{ Class: SmallInteger }" |

    val := anInteger.
    (val between:0 and:9) ifTrue:[
        ^ Character value:(val + ($0 codePoint))
    ].
    (val between:10 and:35) ifTrue:[
        ^ Character value:(val + ($A codePoint - 10))
    ].
    ^self error:'value not in range 0 to 35'
!

utf8DecodeFrom:aStream
    "read and return a single unicode character from an UTF8 encoded stream"

    |fetchNext c1 c2 c3 c4 c5 ascii|

    c1 := aStream next.
    ascii := c1 codePoint.
    ascii <= 16r7F ifTrue:[
        "/ 0xxxxxxx - 7 bits   
        ^ c1.
    ].

    (ascii bitAnd:2r11000000) == 2r10000000 ifTrue:[
        "/ out of sync (got an intermediate character)
        self error:'out of sync'.
        ^ c1.
    ].

    fetchNext := [  |ch|

                    ch := aStream next.
                    (ch codePoint bitAnd:2r11000000) == 2r10000000 ifFalse:[
                        "/ followup chars must have 2r10 in high bits
                        self error:'bad encoding'.
                    ].
                    ch
                 ].

    (ascii bitAnd:2r11100000) == 2r11000000 ifTrue:[
        "/ 110xxxxx 10xxxxxx - 11 bits
        c2 := fetchNext value.
        ascii := c1 codePoint bitAnd:16r1F.
        ascii := (ascii bitShift:6) bitOr:(c2 codePoint bitAnd:16r3F).
        ascii <= 16r7F ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].
    (ascii bitAnd:2r11110000) == 2r11100000 ifTrue:[
        "/ 1110xxxx 10xxxxxx 10xxxxxx - 16 bits
        c2 := fetchNext value.
        c3 := fetchNext value.
        ascii := c1 codePoint bitAnd:16r0F.
        ascii := (ascii bitShift:6) bitOr:(c2 codePoint bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c3 codePoint bitAnd:16r3F).
        ascii <= 16r7FF ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].

    "/ notice: currently, characters can only have 16bit encoding;
    "/ therefore the following will raise a runtime exception,

    (ascii bitAnd:2r11111000) == 2r11110000 ifTrue:[
        "/ 11110xxx 10xxxxxx 10xxxxxx 10xxxxxx - 21 bits
        c2 := fetchNext value.
        c3 := fetchNext value.
        c4 := fetchNext value.
        ascii := c1 codePoint bitAnd:16r07.
        ascii := (ascii bitShift:6) bitOr:(c2 codePoint bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c3 codePoint bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c4 codePoint bitAnd:16r3F).
        ascii <= 16rFFFF ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].

    (ascii bitAnd:2r11111100) == 2r11111000 ifTrue:[
        "/ 111110xx 10xxxxxx 10xxxxxx 10xxxxxx 10xxxxxx - 26 bits
        c2 := fetchNext value.
        c3 := fetchNext value.
        c4 := fetchNext value.
        c5 := fetchNext value.
        ascii := c1 codePoint bitAnd:16r03.
        ascii := (ascii bitShift:6) bitOr:(c2 codePoint bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c3 codePoint bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c4 codePoint bitAnd:16r3F).
        ascii := (ascii bitShift:6) bitOr:(c5 codePoint bitAnd:16r3F).
        ascii <= 16r1FFFFF ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].

    (ascii bitAnd:2r11111110) == 2r11111100 ifTrue:[
        "/ 1111110x ... 10xxxxxx - any number of bits
        ascii := c1 codePoint bitAnd:16r01.

        c2 := aStream peek.
        [ (c2 codePoint bitAnd:2r11000000) == 2r10000000 ] whileTrue:[
            ascii := (ascii bitShift:6) bitOr:(c2 codePoint bitAnd:16r3F).
            aStream next.
        ].
        ascii <= 16r3FFFFFF ifTrue:[
            self error:'invalid encoding'.
        ].
        ^ Character value:ascii        
    ].

    self error:'bad encoding'.
    ^ c1

    "
      Character utf8DecodeFrom:'a' readStream     
      Character utf8DecodeFrom:#[195 188] asString readStream   
    "                                            

    "test:

      |utf8Encoding original readBack|

      1 to:16rFFFF do:[:ascii |
        original := Character value:ascii.
        utf8Encoding := original asString utf8Encoded.
        readBack := Character utf8DecodeFrom:(utf8Encoding readStream).
        readBack codePoint = ascii ifFalse:[
            self halt
        ]
      ]
    "
!

value:anInteger
    "return a character with asciivalue anInteger"

%{  /* NOCONTEXT */

    INT __codePoint;

    if (__isSmallInteger(anInteger)) {
        __codePoint = __smallIntegerVal(anInteger);
        if ((unsigned)(__codePoint) <= MAX_IMMEDIATE_CHARACTER /* (__codePoint >= 0) && (__codePoint <= 255) */) {
            RETURN ( __MKCHARACTER(__codePoint) );
        }
    }
%}.
    (anInteger between:0 and:(CharacterTable size - 1)) ifTrue:[
        ^ CharacterTable at:(anInteger + 1)
    ].
    (anInteger between:16r100 and:16r3FFFFFFF) ifTrue:[
        ^ super basicNew setCodePoint:anInteger
    ].
    "
     a characters ascii-code must be 0..16r3FFFFFFF.
     (i.e. only characters with up-to 30 bits are allowed.)
    "
    self error:'invalid ascii code for character'
! !

!Character class methodsFor:'Compatibility-ANSI'!

codePoint:asciiValue
    ^ self value:asciiValue
! !

!Character class methodsFor:'accessing untypeable characters'!

endOfInput
    "Answer the Character representing ctrl-d ."
    
    ^ self value:4
!

leftParenthesis
    "Answer the Character representing a left parenthesis."
    
    ^ self value:40
!

period
    "Answer the Character representing a carriage period."
    
    ^ self value:46
!

poundSign
    "Answer the Character representing a pound sign."
    
    ^ self value:35
!

rightParenthesis
    "Answer the Character representing a right parenthesis."
    
    ^ self value:41
! !

!Character class methodsFor:'constants'!

backspace
    "return the backspace character"

    ^ Character value:8
!

bell
    "return the bell character"

    ^ Character value:7
!

cr
    "return the lineEnd character 
     - actually (in unix) this is a newline character"

    ^ Character value:10
!

del 
    "return the delete character"

    ^ Character value:16r7F 
!

doubleQuote
    "return the double-quote character"

    ^ Character value:34
!

esc
    "return the escape character"

    ^ Character value:27
!

euro
    "The Euro currency sign (notice: not all fonts support it).
     Thats the ascii-character code; the Unicode encoding is 0x8364"

    ^ Character value: 16r20AC

    "
     Transcript font:(Font family:'courier' size:18 encoding:'iso10646-1').
     Transcript showCR:Character euro asString
    "
    "
     0 to:255 do:[:i |
        Transcript 
            show:'| '; show:((i printStringRadix:16) leftPaddedTo:2);
            show:' | '; show:(i printStringPaddedTo:3);
            show:' | '; show:(Character value:i);
            cr.
     ]
    "
!

excla
    "return the exclamation-mark character"
    ^ $!!
!

ff
    "return the form-feed character"

    ^ Character value:12
!

lf
    "return the newline/linefeed character"

    ^ Character value:10
!

linefeed
    "squeak compatibility: return the newline/linefeed character"

    ^ Character value:10
!

maxValue 
    "return the maximum asciiValue a character can have"

    ^ 16rFFFF
!

newPage
    "return the form-feed character"

    ^ Character value:12
!

nl
    "return the newline character"

    ^ Character value:10
!

null
    ^ Character value:0
!

quote
    "return the single-quote character"

    ^ Character value:39
!

return 
    "return the return character.
     In ST/X, this is different from cr - for Unix reasons."

    ^ Character value:13 
!

space
    "return the blank character"

    ^ Character value:32
!

tab
    "return the tabulator character"

    ^ Character value:9
! !

!Character class methodsFor:'primitive input'!

fromUser
    "return a character from the keyboard (C's standard input stream)
     - this should only be used for emergency evaluators and the like."

%{  /* NOCONTEXT */
    int c;

    c = getchar();
    if (c < 0) {
        RETURN (nil);
    }
    RETURN ( __MKCHARACTER(c & 0xFF) );
%}.
    ^ Stdin next
! !

!Character class methodsFor:'queries'!

allCharacters
    "added for squeak compatibility: return a collection of all chars"

    ^ CharacterTable
        
    "
     Character allCharacters 
    "
!

hasSharedInstances
    "return true if this class has shared instances, that is, instances
     with the same value are identical.
     False is returned here, only redefined in classes which have unified
     instances (or should be treated so)."

    ^ true


!

isBuiltInClass
    "return true if this class is known by the run-time-system.
     Here, true is returned for myself, false for subclasses."

    ^ self == Character

    "Modified: 23.4.1996 / 15:56:39 / cg"
!

isLegalUnicodeCodePoint:anInteger
    "answer true, if anInteger is a valid unicode code point"

    "Range 16rD800 - 16rDFFF is reserved for the 
     lower and upper substitution page for UCS-16"
    (anInteger >= 16rD800) ifTrue:[
        (anInteger <= 16rDFFF) ifTrue:[
            ^ false.
        ].
        (anInteger == 16rFFFE) ifTrue:[
            ^ false.
        ].
        (anInteger == 16rFFFF) ifTrue:[
            ^ false.
        ].
    ].
    ^ true
!

separators
    "added for squeak compatibility: return a collection of separator chars"

    ^ Array 
        with:Character space
        with:Character return
        "/ with:Character cr
        with:Character tab
        with:Character lf
        with:Character ff
        
    "
     Character separators
    "
! !

!Character methodsFor:'Compatibility-Dolphin'!

isAlphaNumeric
    "return true, if I am a letter or a digit
     Please use isLetterOrDigit for compatibility reasons (which is ANSI)."

    ^ self isLetterOrDigit
!

isAlphabetic
    "return true, if I am a letter.
     Please use isLetter for compatibility reasons (which is ANSI)."

    ^ self isLetter
!

isControl
    "return true if I am a control character (i.e. ascii value < 32)"

    ^ self isControlCharacter
!

isHexDigit
    "return true if I am a valid hexadecimal digit"

    ^ '0123456789abcdefABCDEF' includes:self

    "
     $a isHexDigit
    "
!

isPunctuation
    "the code below is not unicode aware"

    ^ (asciivalue between:16r21 and:16r40)
      or:[ (asciivalue between:16r5B and:16r60)
      or:[ (asciivalue between:123 and:126)
      or:[ (asciivalue between:161 and:191)
      or:[ (asciivalue == 215 )
      or:[ (asciivalue == 247 ) ]]]]]
! !

!Character methodsFor:'accessing'!

codePoint
    "return the codePoint of myself.
     Traditionally, this was named 'asciiValue'; 
     however, characters are not limited to 8bit characters."

    ^ asciivalue
!

instVarAt:index put:anObject
    "catch instvar access - asciivalue may not be changed"

    self error:'Characters may not be modified'
! !

!Character methodsFor:'arithmetic'!

+ aMagnitude
    "Return the Character that is <aMagnitude> higher than the receiver. 
     Wrap if the resulting value is not a legal Character value. (JS)"

    ^ Character value:(asciivalue + aMagnitude asInteger \\ 16rFFFF)

    "
     $A + 5
    "

    "Modified: 27.6.1996 / 12:34:51 / cg"
!

- aMagnitude
    "Return the Character that is <aMagnitude> lower than the receiver.  
     Wrap if the resulting value is not a legal Character value. (JS)
     claus: 
        modified to return the difference as integer, if the argument
        is another character. If the argument is a number, a character is
        returned."

    aMagnitude isCharacter ifTrue:[
        ^ asciivalue - aMagnitude asInteger
    ].
    ^ Character value:(asciivalue - aMagnitude asInteger \\ 16rFFFF)

    "
     $z - $a  
     $d - 3
    "

    "Modified: 27.6.1996 / 12:35:34 / cg"
!

// aMagnitude
    "Return the Character who's value is the receiver divided by <aMagnitude>. 
     Wrap if the resulting value is not a legal Character value. (JS)"

    ^ Character value:(asciivalue // aMagnitude asInteger \\ 16rFFFF)
!

\\ aMagnitude
    "Return the Character who's value is the receiver modulo <aMagnitude>.  
     Wrap if the resulting value is not a legal Character value. (JS)"

    ^ Character value:(asciivalue \\ aMagnitude asInteger \\ 16rFFFF)
! !

!Character methodsFor:'binary storage'!

hasSpecialBinaryRepresentation
    "return true, if the receiver has a special binary representation"

    ^ true
!

storeBinaryOn:stream manager:manager
    "store a binary representation of the receiver on stream;
     redefined, since single-byte characters are stored more compact
     with a special type-code followed by the asciiValue."

    (asciivalue < 256) ifTrue:[
	stream nextPut:manager codeForCharacter; nextPut:asciivalue.
    ] ifFalse:[
	stream nextPut:manager codeForTwoByteCharacter.
	stream nextPutShort:asciivalue MSB:true
    ]
! !

!Character methodsFor:'comparing'!

< aMagnitude
    "return true, if the arguments asciiValue is greater than the receiver's"

    ^ (asciivalue < aMagnitude asInteger)
!

<= aMagnitude
    "return true, if the arguments asciiValue is greater or equal to the receiver's"

    ^ (asciivalue <= aMagnitude asInteger)
!

= aCharacter
    "return true, if the argument, aCharacter is the same character
     Redefined to take care of character sizes > 8bit."

    self == aCharacter ifTrue:[^ true].
    aCharacter isCharacter ifFalse:[^ false].
    ^ asciivalue = aCharacter codePoint

    "
        $A = 65
        $A = 43
    "
!

> aMagnitude
    "return true, if the arguments asciiValue is less than the receiver's"

    ^ (asciivalue > aMagnitude asInteger)
!

>= aMagnitude
    "return true, if the arguments asciiValue is less or equal to the receiver's"

    ^ (asciivalue >= aMagnitude asInteger)
!

hash
    "return an integer useful for hashing"

    ^ asciivalue
!

identityHash
    "return an integer useful for hashing on identity"

%{  
    INT __codePoint;

    __codePoint = __smallIntegerVal(__INST(asciivalue));

    if (__codePoint <= MAX_IMMEDIATE_CHARACTER) {
        RETURN ( __mkSmallInteger(__codePoint + 4096) );
    }
%}.

    ^ super identityHash

    "
      $a identityHash.
      (Character value:1234) identityHash
    "
!

sameAs:aCharacter
    "return true, if the argument, aCharacter is the same character,
     ignoring case differences."

    self == aCharacter ifTrue:[^ true].
    ^ self asLowercase = aCharacter asLowercase
!

~= aCharacter
    "return true, if the argument, aCharacter is not the same character
     Redefined to take care of character sizes > 8bit."

    self == aCharacter ifTrue:[^ false].
    aCharacter isCharacter ifFalse:[^ true].
    ^ (asciivalue ~~ aCharacter codePoint)
! !

!Character methodsFor:'converting'!

asCharacter
    "usually sent to integers, but redefined here to allow integers
     and characters to be used commonly without a need for a test."

    ^ self

    "
     32 asCharacter  
    "
!

asInteger
    "the same as #codePoint.
     Use #asInteger, if you need protocol compatibility with Numbers etc..
     Use #codePoint in any other case for better stc optimization"

    ^ asciivalue
!

asLowercase
    "return a character with same letter as the receiver, but in lowercase.
     Returns the receiver if it is already lowercase or if there is no lowercase equivalent.
     CAVEAT:
        for now, this method is only correct for unicode characters up to u+1d6ff (Unicode3.1).
        (which is more than mozilla does, btw. ;-)"

%{
    static int 
    __mapping[] = {
    /* From    To             Every   Diff   */ 
       0x0041, ((0x19 << 8) | 0x01), 0x0020  ,
       0x00c0, ((0x16 << 8) | 0x01), 0x0020  ,
       0x00d8, ((0x06 << 8) | 0x01), 0x0020  ,
       0x0100, ((0x2e << 8) | 0x02), 0x0001  ,
       0x0130, ((0x00 << 8) | 0x00), -199   ,
       0x0132, ((0x04 << 8) | 0x02), 0x0001  ,
       0x0139, ((0x0e << 8) | 0x02), 0x0001  ,
       0x014a, ((0x2c << 8) | 0x02), 0x0001  ,
       0x0178, ((0x00 << 8) | 0x00), -121   ,
       0x0179, ((0x04 << 8) | 0x02), 0x0001  ,
       0x0181, ((0x00 << 8) | 0x00), 0x00d2  ,
       0x0182, ((0x02 << 8) | 0x02), 0x0001  ,
       0x0186, ((0x00 << 8) | 0x00), 0x00ce  ,
       0x0187, ((0x00 << 8) | 0x00), 0x0001  ,
       0x0189, ((0x01 << 8) | 0x01), 0x00cd  ,
       0x018b, ((0x00 << 8) | 0x00), 0x0001  ,
       0x018e, ((0x00 << 8) | 0x00), 0x004f  ,
       0x018f, ((0x00 << 8) | 0x00), 0x00ca  ,
       0x0190, ((0x00 << 8) | 0x00), 0x00cb  ,
       0x0191, ((0x00 << 8) | 0x00), 0x0001  ,
       0x0193, ((0x00 << 8) | 0x00), 0x00cd  ,
       0x0194, ((0x00 << 8) | 0x00), 0x00cf  ,
       0x0196, ((0x00 << 8) | 0x00), 0x00d3  ,
       0x0197, ((0x00 << 8) | 0x00), 0x00d1  ,
       0x0198, ((0x00 << 8) | 0x00), 0x0001  ,
       0x019c, ((0x00 << 8) | 0x00), 0x00d3  ,
       0x019d, ((0x00 << 8) | 0x00), 0x00d5  ,
       0x019f, ((0x00 << 8) | 0x00), 0x00d6  ,
       0x01a0, ((0x04 << 8) | 0x02), 0x0001  ,
       0x01a6, ((0x00 << 8) | 0x00), 0x00da  ,
       0x01a7, ((0x00 << 8) | 0x00), 0x0001  ,
       0x01a9, ((0x00 << 8) | 0x00), 0x00da  ,
       0x01ac, ((0x00 << 8) | 0x00), 0x0001  ,
       0x01ae, ((0x00 << 8) | 0x00), 0x00da  ,
       0x01af, ((0x00 << 8) | 0x00), 0x0001  ,
       0x01b1, ((0x01 << 8) | 0x01), 0x00d9  ,
       0x01b3, ((0x02 << 8) | 0x02), 0x0001  ,
       0x01b7, ((0x00 << 8) | 0x00), 0x00db  ,
       0x01b8, ((0x04 << 8) | 0x04), 0x0001  ,
       0x01c4, ((0x00 << 8) | 0x00), 0x0002  ,
       0x01c5, ((0x00 << 8) | 0x00), 0x0001  ,
       0x01c7, ((0x00 << 8) | 0x00), 0x0002  ,
       0x01c8, ((0x00 << 8) | 0x00), 0x0001  ,
       0x01ca, ((0x00 << 8) | 0x00), 0x0002  ,
       0x01cb, ((0x10 << 8) | 0x02), 0x0001  ,
       0x01de, ((0x10 << 8) | 0x02), 0x0001  ,
       0x01f1, ((0x00 << 8) | 0x00), 0x0002  ,
       0x01f2, ((0x02 << 8) | 0x02), 0x0001  ,
       0x01f6, ((0x00 << 8) | 0x00), -97   ,
       0x01f7, ((0x00 << 8) | 0x00), -56   ,
       0x01f8, ((0x26 << 8) | 0x02), 0x0001  ,
#ifndef UNICODE_3_2
       0x0220, ((0x00 << 8) | 0x00), -130  ,          /* Unicode4.0 - not in X fonts - sigh */
#endif
       0x0222, ((0x10 << 8) | 0x02), 0x0001  ,
       0x0386, ((0x00 << 8) | 0x00), 0x0026  ,
       0x0388, ((0x02 << 8) | 0x01), 0x0025  ,
       0x038c, ((0x00 << 8) | 0x00), 0x0040  ,
       0x038e, ((0x01 << 8) | 0x01), 0x003f  ,
       0x0391, ((0x10 << 8) | 0x01), 0x0020  ,
       0x03a3, ((0x08 << 8) | 0x01), 0x0020  ,
#ifndef UNICODE_3_2
       0x03d8, ((0x00 << 8) | 0x00), 1  ,             /* Unicode4.0 - not in X fonts - sigh */
#endif
       0x03da, ((0x14 << 8) | 0x02), 0x0001  ,
       0x03f4, ((0x00 << 8) | 0x00), -60   ,
#ifndef UNICODE_3_2
       0x03f7, ((0x03 << 8) | 0x03), 1  ,             /* Unicode4.0 - not in X fonts - sigh */
       0x03f9, ((0x00 << 8) | 0x00), -7 ,             /* Unicode4.0 - not in X fonts - sigh */
#endif
       0x0400, ((0x0f << 8) | 0x01), 0x0050  ,
       0x0410, ((0x1f << 8) | 0x01), 0x0020  ,
       0x0460, ((0x20 << 8) | 0x02), 0x0001  ,
#ifndef UNICODE_3_2
       0x048a, ((0x00 << 8) | 0x00), 1 ,              /* Unicode4.0 - not in X fonts - sigh */
#endif
       0x048c, ((0x32 << 8) | 0x02), 0x0001  ,
#ifdef UNICODE_3_2
       0x04c1, ((0x02 << 8) | 0x02), 0x0001  ,
#else
       0x04c1, ((0x04 << 8) | 0x02), 0x0001  ,        /* Unicode4.0 - not in X fonts - sigh */
#endif
#ifdef UNICODE_3_2
       0x04c7, ((0x04 << 8) | 0x04), 0x0001  ,
#else
       0x04c7, ((0x04 << 8) | 0x02), 0x0001  ,
       0x04cd, ((0x00 << 8) | 0x00), 0x0001  ,
#endif
       0x04d0, ((0x24 << 8) | 0x02), 0x0001  ,
       0x04f8, ((0x00 << 8) | 0x00), 0x0001  ,
#ifndef UNICODE_3_2
       0x0500, ((0x0E << 8) | 0x02), 1  ,
#endif
       0x0531, ((0x25 << 8) | 0x01), 0x0030  ,
       0x1e00, ((0x94 << 8) | 0x02), 0x0001  ,
       0x1ea0, ((0x58 << 8) | 0x02), 0x0001  ,
       0x1f08, ((0x07 << 8) | 0x01), -8   ,
       0x1f18, ((0x05 << 8) | 0x01), -8   ,
       0x1f28, ((0x07 << 8) | 0x01), -8   ,
       0x1f38, ((0x07 << 8) | 0x01), -8   ,
       0x1f48, ((0x05 << 8) | 0x01), -8   ,
       0x1f59, ((0x06 << 8) | 0x02), -8   ,
       0x1f68, ((0x07 << 8) | 0x01), -8   ,
       0x1f88, ((0x07 << 8) | 0x01), -8   ,
       0x1f98, ((0x07 << 8) | 0x01), -8   ,
       0x1fa8, ((0x07 << 8) | 0x01), -8   ,
       0x1fb8, ((0x01 << 8) | 0x01), -8   ,
       0x1fba, ((0x01 << 8) | 0x01), -74  ,
       0x1fbc, ((0x00 << 8) | 0x00), -9   ,
       0x1fc8, ((0x03 << 8) | 0x01), -86  ,
       0x1fcc, ((0x00 << 8) | 0x00), -9   ,
       0x1fd8, ((0x01 << 8) | 0x01), -8   ,
       0x1fda, ((0x01 << 8) | 0x01), -100 ,
       0x1fe8, ((0x01 << 8) | 0x01), -8   ,
       0x1fea, ((0x01 << 8) | 0x01), -112 ,
       0x1fec, ((0x00 << 8) | 0x00), -7   ,
       0x1ff8, ((0x01 << 8) | 0x01), -128 ,
       0x1ffa, ((0x01 << 8) | 0x01), -126 ,
       0x1ffc, ((0x00 << 8) | 0x00), -9   ,
       0x2126, ((0x00 << 8) | 0x00), -7517   ,
       0x212a, ((0x00 << 8) | 0x00), -8383   ,
       0x212b, ((0x00 << 8) | 0x00), -8262   ,
       0x2160, ((0x0f << 8) | 0x01), 0x0010  ,
       0x24b6, ((0x19 << 8) | 0x01), 0x001a  ,
       0xff21, ((0x19 << 8) | 0x01), 0x0020  ,  
       0x10400, ((0x27 << 8) | 0x01), 0x0028   
    };

    REGISTER unsigned __codePoint;
    REGISTER int *  __p;

    __codePoint = __intVal(__INST(asciivalue));
    for (__p = __mapping; (char *)__p < ((char *)__mapping) + sizeof(__mapping); __p += 3) {
        unsigned rangeStart, rangeSize, rangeEnd, mod;

        rangeStart = (unsigned)__p[0];
        if (__codePoint < rangeStart) break;

        rangeSize = ((unsigned)__p[1]) >> 8;
        rangeEnd = rangeStart + rangeSize;
        if (__codePoint <= rangeEnd) {
            mod = __p[1] & 0xFF;       
            if ((mod == 0) || (((__codePoint - rangeStart) % mod) == 0)) {
                OBJ newChar;
                unsigned newCodePoint;

                newCodePoint = __codePoint + __p[2];
                if (newCodePoint <= MAX_IMMEDIATE_CHARACTER) {
                    RETURN (__MKCHARACTER(newCodePoint)) ;
                }
                newChar = __MKUCHARACTER(newCodePoint) ;
                if (newChar == nil) goto allocationError;
                RETURN (newChar) ;
            }
        }
    }
    RETURN (self);
allocationError: ;
%}.
    ^ ObjectMemory allocationFailureSignal raise.

    "
     $A asLowercase 
     $a asLowercase 
     (Character value:16r01F5) asUppercase asLowercase 
     (Character value:16r0205) asUppercase asLowercase 
     (Character value:16r03B1) asUppercase asLowercase 
     (Character value:16r1E00) asLowercase  
    "    
!

asString
    "return a string of len 1 with myself as contents"

%{  /* NOCONTEXT */
    char buffer[2];
    OBJ s;
    unsigned val;

    val = __intVal(_characterVal(self));
    if (val <= 0xFF) {
        buffer[0] = (char) val;
        buffer[1] = '\0';
        s = __MKSTRING_L(buffer, 1);
        if (s != nil) {
            RETURN (s);
        }
    }
    if (val <= 0xFFFF) {
        s = __MKEMPTYUSTRING(1);
        if (s != nil) {
            __Unicode16StringInstPtr(s)->s_element[0] = val;
            RETURN (s);
        }
    }
%}.
    asciivalue > 16rFF ifTrue:[
        asciivalue > 16rFFFF ifTrue:[
            ^ (Unicode32String new:1) at:1 put:self; yourself
        ].
        ^ (Unicode16String new:1) at:1 put:self; yourself
    ].

    ^ (String new:1) at:1 put:self; yourself.

    "
     (Character value:16rB5) asString     
     (Character value:16r1B5) asString     
    "
!

asSymbol
    "return a unique symbol which prints like I print"

    ^ Symbol internCharacter:self
!

asTitlecase
    "return a character with same letter as the receiver, but in titlecase.
     Returns the receiver if it is already titlecase or if there is no titlecase equivalent."

    "
     For example, in Unicode, character U+01F3 is LATIN SMALL LETTER DZ.  
     (Let us write this compound character using ASCII as 'dz'.) 
     This character uppercases to character U+01F1, LATIN CAPITAL LETTER DZ.  
     (Which is basically 'DZ'.) 
     But it titlecases to to character U+01F2, LATIN CAPITAL LETTER D WITH SMALL LETTER Z. 
     (Which we can write 'Dz'.)

      character uppercase titlecase
      --------- --------- ---------
      dz        DZ        Dz
    "
    |ch|

%{
    static unsigned short
    __mapping[] = {
       0x01F1,    0x01F2, 
       0x01F3,    0x01F2, 
       0x01C4,    0x01C5, 
       0x01C6,    0x01C5, 
       0x01C7,    0x01C8, 
       0x01C9,    0x01C8, 
       0x01CA,    0x01CB, 
       0x01CC,    0x01CB, 
    };

    REGISTER unsigned __codePoint;
    REGISTER short *__p;

    __codePoint = __intVal(__INST(asciivalue));
    for (__p = __mapping; (char *)__p < ((char *)__mapping) + sizeof(__mapping); __p += 2) {
        if ((__codePoint == __p[0]) || (__codePoint == __p[1])) {
            short newCodePoint;
            OBJ newChar;

            newCodePoint = __p[1];
            if (newCodePoint == __codePoint) {
                RETURN (self);
            }
            if (newCodePoint <= MAX_IMMEDIATE_CHARACTER) {
                RETURN (__MKCHARACTER(newCodePoint)) ;
            }
            newChar = __MKUCHARACTER(newCodePoint) ;
            if (newChar == nil) goto getOutOfHere;
            RETURN (newChar) ;
        }
    }
    ch = self;
getOutOfHere: ;
%}.
    ch notNil ifTrue:[
        ^ ch asUppercase.
    ].

    ^ ObjectMemory allocationFailureSignal raise.

    "
     $A asTitlecase 
     $a asTitlecase 
     (Character value:16r01F1) asTitlecase 
    "    
!

asUnicodeString
    "return a unicode string of len 1 with myself as contents.
     This will vanish, as we now (rel5.2.x) use Unicode as default."

    asciivalue > 16rFFFF ifTrue:[
        ^ (Unicode32String new:1) at:1 put:self; yourself.
    ].
    ^ (Unicode16String new:1) at:1 put:self; yourself.
!

asUppercase
    "return a character with same letter as the receiver, but in uppercase.
     Returns the receiver if it is already uppercase or if there is no uppercase equivalent.
     CAVEAT:
        for now, this method is only correct for unicode characters up to u+1d6ff (Unicode3.1).
        (which is more than mozilla does, btw. ;-)"

%{
    static int
    __mapping[] = {
    /* From    To             Every   Diff   */ 
       0x0061, ((0x19 << 8) | 0x01), -32  ,
       0x00b5, ((0x00 << 8) | 0x3b), 0x02e7  ,
       0x00e0, ((0x16 << 8) | 0x01), -32   ,
       0x00f8, ((0x06 << 8) | 0x01), -32   ,
       0x00ff, ((0x00 << 8) | 0x01), 0x0079  ,
       0x0101, ((0x2e << 8) | 0x02), -1   ,
       0x0131, ((0x00 << 8) | 0x02), -232  ,
       0x0133, ((0x04 << 8) | 0x02), -1   ,
       0x013a, ((0x0e << 8) | 0x02), -1   ,
       0x014b, ((0x2c << 8) | 0x02), -1   ,
       0x017a, ((0x04 << 8) | 0x02), -1   ,
       0x017f, ((0x00 << 8) | 0x01), -300  ,
       0x0183, ((0x02 << 8) | 0x02), -1   ,
       0x0188, ((0x04 << 8) | 0x04), -1   ,
       0x0192, ((0x00 << 8) | 0x06), -1   ,
       0x0195, ((0x00 << 8) | 0x03), 0x0061  ,
#ifndef UNICODE_3_2
       0x0199, ((0x04 << 8) | 0x08), -1   ,
       0x019e, ((0x00 << 8) | 0x00), 130  ,          /* Unicode4.0 - not in X fonts - sigh */
#endif
       0x0199, ((0x08 << 8) | 0x08), -1   ,
       0x01a3, ((0x02 << 8) | 0x02), -1   ,
       0x01a8, ((0x05 << 8) | 0x05), -1   ,
       0x01b0, ((0x04 << 8) | 0x04), -1   ,
       0x01b6, ((0x03 << 8) | 0x03), -1   ,
       0x01bd, ((0x00 << 8) | 0x04), -1   ,
       0x01bf, ((0x00 << 8) | 0x02), 0x0038  ,
       0x01c5, ((0x00 << 8) | 0x06), -1   ,
       0x01c6, ((0x00 << 8) | 0x01), -2   ,
       0x01c8, ((0x00 << 8) | 0x02), -1   ,
       0x01c9, ((0x00 << 8) | 0x01), -2   ,
       0x01cb, ((0x00 << 8) | 0x02), -1   ,
       0x01cc, ((0x00 << 8) | 0x01), -2   ,
       0x01ce, ((0x0e << 8) | 0x02), -1   ,
       0x01dd, ((0x00 << 8) | 0x01), -79   ,
       0x01df, ((0x10 << 8) | 0x02), -1   ,
       0x01f2, ((0x00 << 8) | 0x03), -1   ,
       0x01f3, ((0x00 << 8) | 0x01), -2   ,
       0x01f5, ((0x04 << 8) | 0x04), -1   ,
       0x01fb, ((0x24 << 8) | 0x02), -1   ,
       0x0223, ((0x10 << 8) | 0x02), -1   ,
       0x0253, ((0x00 << 8) | 0x20), -210   ,
       0x0254, ((0x00 << 8) | 0x01), -206   ,
       0x0256, ((0x01 << 8) | 0x01), -205   ,
       0x0259, ((0x00 << 8) | 0x02), -202   ,
       0x025b, ((0x00 << 8) | 0x02), -203   ,
       0x0260, ((0x00 << 8) | 0x05), -205   ,
       0x0263, ((0x00 << 8) | 0x03), -207   ,
       0x0268, ((0x00 << 8) | 0x05), -209   ,
       0x0269, ((0x06 << 8) | 0x06), -211   ,
       0x0272, ((0x00 << 8) | 0x03), -213   ,
       0x0275, ((0x00 << 8) | 0x03), -214   ,
       0x0280, ((0x03 << 8) | 0x03), -218   ,
       0x0288, ((0x00 << 8) | 0x05), -218   ,
       0x028a, ((0x01 << 8) | 0x01), -217   ,
       0x0292, ((0x00 << 8) | 0x07), -219   ,
       0x0345, ((0x00 << 8) | 0xb3), 0x0054  ,
       0x03ac, ((0x00 << 8) | 0x67), -38   ,
       0x03ad, ((0x02 << 8) | 0x01), -37   ,
       0x03b1, ((0x10 << 8) | 0x01), -32   ,
       0x03c2, ((0x00 << 8) | 0x01), -31   ,
       0x03c3, ((0x08 << 8) | 0x01), -32   ,
       0x03cc, ((0x00 << 8) | 0x01), -64   ,
       0x03cd, ((0x01 << 8) | 0x01), -63   ,
       0x03d0, ((0x00 << 8) | 0x02), -62   ,
       0x03d1, ((0x00 << 8) | 0x01), -57   ,
       0x03d5, ((0x00 << 8) | 0x04), -47   ,
       0x03d6, ((0x00 << 8) | 0x01), -54   ,
#ifndef UNICODE_3_2
       0x03d9, ((0x00 << 8) | 0x00), -1  ,             /* Unicode4.0 - not in X fonts - sigh */
#endif
       0x03db, ((0x14 << 8) | 0x02), -1    ,
       0x03f0, ((0x00 << 8) | 0x01), -86   ,
       0x03f1, ((0x00 << 8) | 0x01), -80   ,
#ifdef UNICODE_3_2
       0x03f2, ((0x00 << 8) | 0x01), -79   ,
#else
       0x03f2, ((0x00 << 8) | 0x00), 7  ,              /* Unicode4.0 - not in X fonts - sigh */
#endif
       0x03f5, ((0x00 << 8) | 0x00), -96   ,
#ifndef UNICODE_3_2
       0x03f8, ((0x03 << 8) | 0x03), -1  ,             /* Unicode4.0 - not in X fonts - sigh */
#endif
       0x0430, ((0x1f << 8) | 0x01), -32   ,
       0x0450, ((0x0f << 8) | 0x01), -80   ,
       0x0461, ((0x20 << 8) | 0x02), -1   ,
#ifndef UNICODE_3_2
       0x048b, ((0x00 << 8) | 0x00), -1 ,              /* Unicode4.0 - not in X fonts - sigh */
#endif
       0x048d, ((0x32 << 8) | 0x02), -1   ,
#ifdef UNICODE_3_2
       0x04c2, ((0x02 << 8) | 0x02), -1   ,
#else
       0x04c2, ((0x04 << 8) | 0x02), -1   ,            /* Unicode4.0 - not in X fonts - sigh */
#endif
#ifdef UNICODE_3_2
       0x04c8, ((0x04 << 8) | 0x04), -1   ,
#else
       0x04c8, ((0x04 << 8) | 0x02), -1   ,
       0x04ce, ((0x00 << 8) | 0x00), -1   ,
#endif
       0x04d1, ((0x24 << 8) | 0x02), -1   ,
       0x04f9, ((0x00 << 8) | 0x04), -1   ,
#ifndef UNICODE_3_2
       0x0501, ((0x0E << 8) | 0x02), -1  ,
#endif
       0x0561, ((0x25 << 8) | 0x01), -48   ,
       0x1e01, ((0x94 << 8) | 0x02), -1   ,
       0x1e9b, ((0x00 << 8) | 0x06), -59   ,
       0x1ea1, ((0x58 << 8) | 0x02), -1   ,
       0x1f00, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f10, ((0x05 << 8) | 0x01), 0x0008  ,
       0x1f20, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f30, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f40, ((0x05 << 8) | 0x01), 0x0008  ,
       0x1f51, ((0x06 << 8) | 0x02), 0x0008  ,
       0x1f60, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f70, ((0x01 << 8) | 0x01), 0x004a  ,
       0x1f72, ((0x03 << 8) | 0x01), 0x0056  ,
       0x1f76, ((0x01 << 8) | 0x01), 0x0064  ,
       0x1f78, ((0x01 << 8) | 0x01), 0x0080  ,
       0x1f7a, ((0x01 << 8) | 0x01), 0x0070  ,
       0x1f7c, ((0x01 << 8) | 0x01), 0x007e  ,
       0x1f80, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1f90, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1fa0, ((0x07 << 8) | 0x01), 0x0008  ,
       0x1fb0, ((0x01 << 8) | 0x01), 0x0008  ,
       0x1fb3, ((0x00 << 8) | 0x02), 0x0009  ,
       0x1fbe, ((0x00 << 8) | 0x0b), -7205   ,
       0x1fc3, ((0x00 << 8) | 0x05), 0x0009  ,
       0x1fd0, ((0x01 << 8) | 0x01), 0x0008  ,
       0x1fe0, ((0x01 << 8) | 0x01), 0x0008  ,
       0x1fe5, ((0x00 << 8) | 0x04), 0x0007  ,
       0x1ff3, ((0x00 << 8) | 0x0e), 0x0009  ,
       0x2170, ((0x0f << 8) | 0x01), -16   ,
       0x24d0, ((0x19 << 8) | 0x01), -26   ,
       0xff41, ((0x19 << 8) | 0x01), -32  ,  
       0x10428, ((0x27 << 8) | 0x01), -40   
    };

    REGISTER unsigned __codePoint;
    REGISTER int *__p;

    __codePoint = __intVal(__INST(asciivalue));
    for (__p = __mapping; (char *)__p < ((char *)__mapping) + sizeof(__mapping); __p += 3) {
        unsigned rangeStart, rangeSize, rangeEnd, mod;

        rangeStart = (unsigned)__p[0];
        if (__codePoint < rangeStart) break;

        rangeSize = ((unsigned)__p[1]) >> 8;
        rangeEnd = rangeStart + rangeSize;
        if (__codePoint <= rangeEnd) {
            mod = __p[1] & 0xFF;       
            if ((mod == 0) || (((__codePoint - rangeStart) % mod) == 0)) {
                OBJ newChar;
                unsigned newCodePoint;

                newCodePoint = __codePoint + __p[2];
                if (newCodePoint <= MAX_IMMEDIATE_CHARACTER) {
                    RETURN (__MKCHARACTER(newCodePoint)) ;
                }
                newChar = __MKUCHARACTER(newCodePoint) ;
                if (newChar == nil) goto allocationError;
                RETURN (newChar) ;
            }
        }
    }
    RETURN (self);
allocationError: ;
%}.
    ^ ObjectMemory allocationFailureSignal raise.

    "
     $A asLowercase 
     $a asUppercase 
     (Character value:16r01F5) asUppercase  
     (Character value:16r0205) asUppercase  
     (Character value:16r03B1) asUppercase  
    "    
!

digitValue
    "return my digitValue for any base"

    |code "{ Class: SmallInteger }" |

    code := asciivalue.
    (code between:($0 codePoint) and:($9 codePoint)) ifTrue:[
        ^ code - $0 codePoint
    ].
    (code between:($a codePoint) and:($z codePoint)) ifTrue:[
        ^ code + (10 - $a codePoint)
    ]. 
    (code between:($A codePoint) and:($Z codePoint)) ifTrue:[
        ^ code + (10 - $A codePoint)
    ]. 

"remove error below for X3J20 conformance ... "
    self error:'bad character'.
" "
    ^ -1
!

literalArrayEncoding
    "encode myself as an array literal, from which a copy of the receiver
     can be reconstructed with #decodeAsLiteralArray."

    ^ self

    "Created: / 27.10.1997 / 14:40:37 / cg"
!

to:aMagnitude
    "Return an Interval over the characters from the receiver to <aMagnitude>. 
     Wrap <aMagnitude> if it is not a legal Character value. (JS)"

    ^ Interval from:self to:(aMagnitude \\ 256)
!

utf8Encoded
    "convert a character from ISO8859-1 to its UTF-8 encoding.
     this returns a String"

    |s|

    asciivalue <= 16r7F ifTrue:[
        ^ self asString.
    ].

    s := WriteStream on:(String new:6).
    self utf8EncodedOn:s.
    ^ s contents
!

utf8EncodedOn:aStream
    "append my UTF-8 representation to the argument, aStream.
     Up to 31 bits can be encoded in up to 6 bytes. 
     However, currently, characters are limited to 16 bits."

    |b1 b2 b3 b4 b5 v|

    asciivalue <= 16r7F ifTrue:[
        aStream nextPut:self.
        ^ self.
    ].

    b1 := Character value:((asciivalue bitAnd:16r3F) bitOr:2r10000000).
    v := asciivalue bitShift:-6.
    v <= 16r1F ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11000000)).
        aStream nextPut:b1.
        ^ self.
    ].
    b2 := Character value:((v bitAnd:16r3F) bitOr:2r10000000).
    v := v bitShift:-6.
    v <= 16r0F ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11100000)).
        aStream nextPut:b2; nextPut:b1.
        ^ self.
    ].
    b3 := Character value:((v bitAnd:16r3F) bitOr:2r10000000).
    v := v bitShift:-6.
    v <= 16r07 ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11110000)).
        aStream nextPut:b3; nextPut:b2; nextPut:b1.
        ^ self.
    ].
    b4 := Character value:((v bitAnd:16r3F) bitOr:2r10000000).
    v := v bitShift:-6.
    v <= 16r03 ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11111000)).
        aStream nextPut:b4; nextPut:b3; nextPut:b2; nextPut:b1.
        ^ self.
    ].
    b5 := Character value:((v bitAnd:16r3F) bitOr:2r10000000).
    v := v bitShift:-6.
    v <= 16r01 ifTrue:[
        aStream nextPut:(Character value:(v bitOr:2r11111100)).
        aStream nextPut:b5; nextPut:b4; nextPut:b3; nextPut:b2; nextPut:b1.
        ^ self.
    ].
    "/ cannot happen - we only support up to 16 bit characters
    self error:'ascii value > 31bit in utf8Encode'.
        
    "
      $a utf8Encoded   
      $ utf8Encoded asByteArray  
      (Character value:16r1fff) utf8Encoded asByteArray      
      (Character value:16rffff) utf8Encoded asByteArray      
      (Character value:16r800) utf8Encoded asByteArray      
    "
! !

!Character methodsFor:'copying'!

copy
    "return a copy of myself
     reimplemented since characters are unique"

     ^ self
!

deepCopyUsing:aDictionary
    "return a deep copy of myself
     reimplemented since characters are immutable"

     ^ self
!

shallowCopy
    "return a shallow copy of myself
     reimplemented since characters are immutable"

     ^ self
!

simpleDeepCopy
    "return a deep copy of myself
     reimplemented since characters are immutable"

     ^ self
! !

!Character methodsFor:'encoding'!

rot13
     "Usenet: from `rotate alphabet 13 places'] 
      The simple Caesar-cypher encryption that replaces each English
      letter with the one 13 places forward or back along the alphabet, 
      so that 'The butler did it!!' becomes 'Gur ohgyre qvq vg!!'
      Most Usenet news reading and posting programs include a rot13 feature. 
      It is used to enclose the text in a sealed wrapper that the reader must choose
      to open -- e.g., for posting things that might offend some readers, or spoilers. 
      A major advantage of rot13 over rot(N) for other N is that it
      is self-inverse, so the same code can be used for encoding and decoding."

    (((self >= $a) and:[ self < $n]) or:[(self >= $A) and:[ self < $N]])
        ifTrue: [ ^ Character value:(self codePoint + 13) ].
    (((self > $m) and:[ self <= $z]) or:[(self > $M) and:[ self <= $Z]])
        ifTrue: [ ^ Character value:(self codePoint - 13) ].
    ^ self

    "
     $h rot13      
     $h rot13 rot13           
     'The butler did it!!' rot13             -> 'Gur ohgyre qvq vg!!'
     'The butler did it!!' rot13 rot13       -> 'The butler did it!!'
    "
! !

!Character methodsFor:'obsolete'!

asciiValue
    "return the asciivalue of myself.
     The name 'asciiValue' is a historic leftover:
        characters are not limited to 8bit characters.
     So the actual value returned is a codePoint (i.e. full potential for 31bit encoding).
     PP has removed this method with 4.1 and providing asInteger instead.
     ANSI defines #codePoint, please use this method"

    <resource:#obsolete>

    ^ asciivalue

    "Modified: 27.6.1996 / 12:34:34 / cg"
! !

!Character methodsFor:'printing & storing'!

displayString
    "return a string used when the receiver is to be displayed
     in an inspector kind-of-thing"

    ^ (self storeString , ' "16r' , (asciivalue printStringRadix:16) , '"')
!

isLiteral
    "return true, if the receiver can be used as a literal constant in ST syntax
     (i.e. can be used in constant arrays)"

    ^ true
!

print
    "print myself on stdout.
     This method does NOT (by purpose) use the stream classes and 
     will therefore work even in case of emergency (but only, if Stdout is nil)."

%{  /* NOCONTEXT */

    if (@global(Stdout) == nil) {
        putchar(__intVal(__INST(asciivalue)));
        RETURN(self);
    }
%}.
    super print
!

printOn:aStream
    "print myself on aStream"

    aStream nextPut:self
!

printString
    "return a string to print me"

    ^ self asString
!

reconstructOn:aStream
    aStream nextPut:$$; nextPut:self

    "Created: 6.2.1996 / 09:46:12 / stefan"
    "Modified: 6.2.1996 / 10:57:35 / stefan"
!

storeOn:aStream
    "store myself on aStream"

    |special|

    (asciivalue between:33 and:127) ifFalse:[
	(self == Character space) ifTrue:[
	    special := 'space'
	] ifFalse:[
	    (self == Character cr) ifTrue:[
		special := 'cr'.
	    ] ifFalse:[
		(self == Character tab) ifTrue:[
		    special := 'tab'.
		] ifFalse:[
		    (self == Character esc) ifTrue:[
			special := 'esc'.
		    ]
		]
	    ]
	].
	special notNil ifTrue:[
	    aStream nextPutAll:'(Character '; nextPutAll:special; nextPut:$).
	    ^ self
	].
	aStream nextPutAll:'(Character value:16r'.
	asciivalue printOn:aStream base:16.
	aStream nextPut:$)
    ] ifTrue:[
	aStream nextPut:$$; nextPut:self
    ]

    "Modified: / 23.2.1996 / 23:27:32 / cg"
    "Modified: / 20.1.1998 / 14:10:46 / stefan"
! !

!Character methodsFor:'private-accessing'!

setCodePoint:anInteger
    "very private - set the codePoint. 
     - use this only for characters with codes > 16rFF.
     DANGER alert: 
        funny things happen, if this is applied to
        one of the shared characters with codePoints 0..255."

    asciivalue := anInteger
! !

!Character methodsFor:'queries'!

bitsPerCharacter
    "return the number of bits I require for storage"

    asciivalue <= 16rFF ifTrue:[^ 8].
    asciivalue <= 16rFFFF ifTrue:[^ 16].
    ^ 32
!

stringSpecies
    "return the type of string that is needed to store me"

    asciivalue <= 16rFF ifTrue:[^ String].
    asciivalue <= 16rFFFF ifTrue:[^ Unicode16String].
    ^ Unicode32String
! !

!Character methodsFor:'testing'!

isCharacter
    "return true,  if the receiver is some kind of character"

    ^ true
!

isControlCharacter
    "return true if I am a control character (i.e. ascii value < 32 or == 16rFF)"

%{  /* NOCONTEXT */
#ifdef NON_ASCII       /* i.e. EBCDIC ;-) */
# error    not yet implemented - fails when compiled
#else
    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if (val < ' ' || val == 0xFF) {
        RETURN ( true );
    }
#endif
    RETURN (false);
%}.

    "
     (Character value:1) isControlCharacter
     $a isControlCharacter
    "
!

isDigit
    "return true, if I am a digit (i.e. $0 .. $9)"

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if ((unsigned)(val - '0') <= ('9' - '0')) {
        RETURN ( true );
    }
    RETURN ( false );
%}.
    ^ asciivalue between:$0 codePoint and:$9 codePoint
!

isDigitRadix:r
    "return true, if I am a digit of a base r number"

    (asciivalue < $0 codePoint) ifTrue:[^ false]. 
    (r > 10) ifTrue:[
        (asciivalue <= $9 codePoint) ifTrue:[
            ^ true
        ].
        ((asciivalue - $a codePoint) between:0 and:(r - 11)) ifTrue:[
            ^ true
        ].
        ^ (asciivalue - $A codePoint) between:0 and:(r - 11)
    ].
    (asciivalue - $0 codePoint) < r ifTrue:[^ true].
    ^ false
!

isEndOfLineCharacter
    "return true if I am a line delimitting character"

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if ((val == '\n')
     || (val == '\r')
     || (val == '\f')) {
        RETURN ( true );
    }
    RETURN (false);
%}.
    ^ asciivalue == 16r0A
      or:[asciivalue == 16r0D
      or:[asciivalue == 16r0C]]

!

isImmediate
    "return true if I am an immediate object
     i.e. I am represented in the pointer itself and
     no real object header/storage is used me.
     For VW compatibility, shared characters (i.e. in the range 0..MAX_IMMEDIATE_CHARACTER) 
     also return true here"
    
%{  /* NOCONTEXT */
    if (__smallIntegerVal(__INST(asciivalue)) <= MAX_IMMEDIATE_CHARACTER) {
        RETURN ( true );
    }
%}.
    ^ false

    "
        $a isImmediate.
        (Character value:255) isImmediate.
        (Character value:256) isImmediate.
        (Character value:1566) isImmediate.
    "
!

isLetter
    "return true, if I am a letter in the 'a'..'z' range.
     Use isNationalLetter, if you are interested in those."

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if ((unsigned)(val - 'a') <= ('z' - 'a')) {
        RETURN ( true );
    }
    if ((unsigned)(val - 'A') <= ('Z' - 'A')) {
        RETURN ( true );
    }
    RETURN ( false );
%}.
    ^ (asciivalue between:($a codePoint) and:($z codePoint))
      or:[(asciivalue between:($A codePoint) and:($Z codePoint))]
!

isLetterOrDigit
    "return true, if I am a letter (a..z or A..Z) or a digit (0..9)
     Use isNationalAlphaNumeric, if you are interested in those."

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    if ((unsigned)(val - 'a') <= ('z' - 'a')) {
        RETURN ( true );
    }
    if ((unsigned)(val - 'A') <= ('Z' - 'A')) {
        RETURN ( true );
    }
    if ((unsigned)(val - '0') <= ('9' - '0')) {
        RETURN ( true );
    }
    RETURN ( false );
%}.
    ^ self isLetter or:[self isDigit]
!

isLowercase
    "return true, if I am a lower-case letter.
     This one does care for national characters.
     Caveat: 
        only returns the correct value for codes up to u+1d6ff (Unicode3.1).
        (which is more than mozilla does, btw. ;-)"

%{  /* NOCONTEXT */

    REGISTER int val;
#define TRUE_IF_ODD(x)      ((x & 1) ? true : false)
#define TRUE_IF_EVEN(x)     ((x & 1) ? false : true)

    val = __intVal(__INST(asciivalue));
    switch (val >> 8) {
        case 0x00:
            if ((unsigned)(val - 'a') <= ('z' - 'a')) {
                RETURN ( true );
            }

            /* iso8859 puts national lower case characters at e0 .. ff */
            if ((val >= 0xDF) && (val <= 0xFF)) {
                if (val != 0xF7) {
                    RETURN(true);
                }
            }
            if (val == 0xAA) RETURN(true);     /* FEMININE ORDINAL INDICATOR (high a-underline) */
            if (val == 0xB5) RETURN(true);     /* MICRO SIGN */
            if (val == 0xBA) RETURN(true);     /* MASCULINE ORDINAL INDICATOR (high o-underline) */
            RETURN (false);

        case 0x01:
            if (val <= 0x0137) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x0148) { RETURN (TRUE_IF_EVEN(val)); } 
            if (val <= 0x0178) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x017E) { RETURN (TRUE_IF_EVEN(val)); } 
            if (val <= 0x0180) { RETURN (true); } 
            if (val < 0x01CD) {
                if (val == 0x0181) { RETURN (false); }
                if (val <= 0x0185) {
                    RETURN (TRUE_IF_ODD(val));
                }
                if (val == 0x0188) { RETURN (true); }
                if (val == 0x018C) { RETURN (true); }
                if (val == 0x018D) { RETURN (true); }
                if (val == 0x0192) { RETURN (true); }
                if (val == 0x0195) { RETURN (true); }
                if (val == 0x0199) { RETURN (true); }
                if (val == 0x019A) { RETURN (true); }
                if (val == 0x019B) { RETURN (true); }
                if (val == 0x019E) { RETURN (true); }
                if (val <= 0x01A0) { RETURN (false); }
                if (val <= 0x01A6) { RETURN (TRUE_IF_ODD(val)); }
                if (val <= 0x01AA) { RETURN (TRUE_IF_EVEN(val)); }
                if (val <= 0x01AE) { RETURN (TRUE_IF_ODD(val)); }
                if (val == 0x01B2) { RETURN (false); }
                if (val <= 0x01B6) { RETURN (TRUE_IF_EVEN(val)); }
                if (val == 0x01B9) { RETURN (true); }
                if (val == 0x01BA) { RETURN (true); }
                if (val == 0x01BD) { RETURN (true); }
                if (val == 0x01BE) { RETURN (true); }
                if (val == 0x01BF) { RETURN (true); }
                if (val == 0x01C6) { RETURN (true); }
                if (val == 0x01C9) { RETURN (true); }
                if (val == 0x01CC) { RETURN (true); }
                RETURN (false);
            } 
            if (val <= 0x01DC) { RETURN (TRUE_IF_EVEN(val)); } 
            if (val <= 0x01EF) { RETURN (TRUE_IF_ODD(val)); } 
            if (val == 0x01F0) { RETURN (true); }
            if (val == 0x01F1) { RETURN (false); }
            if (val == 0x01F2) { RETURN (false); }
            if (val == 0x01F3) { RETURN (true); }
            if (val <= 0x01F6) { RETURN (TRUE_IF_ODD(val)); } 
            if (val == 0x01F7) { RETURN (false); }
            RETURN (TRUE_IF_ODD(val));

        case 0x02:
            if (val <= 0x0233) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x0236) { RETURN (true); } 
            if (val < 0x0250) { RETURN (false); } 
            if (val < 0x02B0) { RETURN (true); } 
            RETURN (false);


        case 0x03:
            if (val == 0x0390) { RETURN (true); } 
            if (val <= 0x03AB) { RETURN (false); } 
            if (val <= 0x03D1) { RETURN (true); } 
            if (val == 0x03D5) { RETURN (true); } 
            if (val == 0x03D6) { RETURN (true); } 
            if (val < 0x03D7) { RETURN (false); } 
            if (val <= 0x03EF) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x03F3) { RETURN (true); } 
            if (val == 0x03F5) { RETURN (true); } 
#ifndef UNICODE_3_2
            if (val == 0x03F8) { RETURN (true); } 
            if (val == 0x03FB) { RETURN (true); } 
#endif
            RETURN (false);

        case 0x04:
            if (val <= 0x042F) { RETURN (false); } 
            if (val <= 0x045F) { RETURN (true); } 
            if (val <= 0x0481) { RETURN (TRUE_IF_ODD(val)); } 
            if (val < 0x048A) { RETURN (false); } 
            if (val <= 0x04C0) { RETURN (TRUE_IF_ODD(val)); }
#ifdef UNICODE_3_2
            if (val == 0x04C5) { RETURN (true); }
#endif
#ifdef UNICODE_3_2
            if (val <= 0x04C8) { RETURN (TRUE_IF_EVEN(val)); }
#else
            if (val <= 0x04CA) { RETURN (TRUE_IF_EVEN(val)); }
            if (val == 0x04CD) { RETURN (false); }
            if (val == 0x04CE) { RETURN (true); }
#endif
            if (val == 0x04CB) { RETURN (false); }
            if (val == 0x04CC) { RETURN (true); }
            RETURN (TRUE_IF_ODD(val));

        case 0x05:
            if (val <= 0x050F) { RETURN (TRUE_IF_ODD(val)); } 
            if (val < 0x0561) { RETURN (false); } 
            if (val <= 0x0587) { RETURN (true); } 
            RETURN (false);

        case 0x1D:
#ifndef UNICODE_3_2
            if (val <= 0x1D2B) { RETURN (true); }
            if (val <= 0x1D61) { RETURN (false); }
            if (val <= 0x1D70) { RETURN (true); }
#endif
            RETURN (false);

        case 0x1E:
            if (val < 0x1E96) { RETURN (TRUE_IF_ODD(val)); }
            if (val <= 0x1E9F) { RETURN (true); }
            RETURN (TRUE_IF_ODD(val));

        case 0x1F:
            if (val <= 0x1F6F) {
                if (val & 0x0008) { RETURN (false); }
                RETURN (true);
            }
            if (val <= 0x1F87) { RETURN (true); }
            if (val < 0x1FB8) {
                if (val & 0x0008) { RETURN (false); }
                RETURN (true);
            }
            if (val == 0x1FBE) { RETURN (true); }
            if (val == 0x1FD4) { RETURN (false); }
            if (val == 0x1FC5) { RETURN (false); }
            if (val == 0x1FD5) { RETURN (false); }
            if (val == 0x1FC1) { RETURN (false); }
            if (val == 0x1FF1) { RETURN (false); }
            if (val == 0x1FC0) { RETURN (false); }
            if (val == 0x1FF0) { RETURN (false); }
            if (((val & 0x000F) >= 0x0000) && ((val & 0x000F) <= 0x0007)) { RETURN (true); }
            RETURN (false);

        case 0x20:
#ifndef UNICODE_3_2
            if (val == 0x2071) { RETURN (true); }
#endif
            if (val == 0x207F) { RETURN (true); }
            RETURN (false);

        case 0x21:
            if (val == 0x210A) { RETURN (true); }
            if (val < 0x210E) { RETURN (false); }
            if (val <= 0x210F) { RETURN (true); }
            if (val == 0x2113) { RETURN (true); }
            if (val == 0x212F) { RETURN (true); }
            if (val == 0x2134) { RETURN (true); }
            if (val == 0x2139) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val == 0x213D) { RETURN (true); }
            if (val <= 0x2145) { RETURN (false); }
            if (val <= 0x2149) { RETURN (true); }
#endif
            RETURN (false);

        case 0xFB:
            if (val <= 0xFB1C) { RETURN (true); }
            RETURN (false);
            
        case 0xFF:
            if ((val >= 0xFF41) && (val <= 0xFF5A)) { RETURN (true); }
            RETURN (false);

        case 0x104:
            if (val <= 0x10427) { RETURN (false); }
            if (val <= 0x1044D) { RETURN (true); }
#ifdef UNICODE_3_2
            if (val <= 0x1044D) { RETURN (true); }
#else
            if (val <= 0x1044F) { RETURN (true); }
#endif
            RETURN (false);

        case 0x1D4:
        case 0x1D5:
        case 0x1D6:
            if (val <= 0x1D419) { RETURN (false); }
            if (val < 0x1D6be) {
                if (((val - 0x1D41A) % 52) <= 25) { RETURN (true); }
                RETURN (false);
            }
            if (val < 0x1D6c2) { RETURN (false); }
            if (val < 0x1D6db) { RETURN (true); }
            if (val == 0x1D6db) { RETURN (false); }
            if (val < 0x1D6e2) { RETURN (true); }
            if (val <= 0x1D6fb) { RETURN (false); }
            RETURN (true);
    }
#undef TRUE_IF_ODD
#undef TRUE_IF_EVEN
    RETURN (false);
%}.
!

isPrintable
    "return true, if the receiver is a useful printable character
     (see fileBrowsers showFile:-method on how it can be used)"

    (asciivalue between:32 and:127) ifTrue:[^ true].
    asciivalue == 12 ifTrue:[^ true].   "/ FF
    asciivalue == 13 ifTrue:[^ true].   "/ CR
    asciivalue == 9 ifTrue:[^ true].    "/ TAB
    asciivalue == 10 ifTrue:[^ true].   "/ NL

    (asciivalue between:16rA0 and:16rBF) ifTrue:[^ true]. "/ ISO-8859
    ^ self isNationalAlphaNumeric

    "Modified: 7.8.1997 / 17:05:24 / cg"
!

isSeparator
    "return true if I am a space, cr, tab, nl, or newPage"

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
#ifndef NON_ASCII       /* i.e. EBCDIC ;-) */
    if (val <= ' ')
#endif
        if ((val == ' ')
         || (val == '\n') 
         || (val == '\t')
         || (val == '\r')
         || (val == '\f')) {
            RETURN ( true );
        }
    RETURN (false);
%}.                                  
    ^ (asciivalue == 16r20)
      or:[asciivalue == 16r0D
      or:[asciivalue == 16r0A
      or:[asciivalue == 16r09
      or:[asciivalue == 16r0C]]]]

!

isUppercase
    "return true, if I am an upper-case letter.
     This one does care for national characters.
     Caveat:   
        only returns the correct value for codes up to u+1d6ff (Unicode3.1).
        (which is more than mozilla does, btw. ;-)"

%{  /* NOCONTEXT */
#define TRUE_IF_ODD(x)      ((x & 1) ? true : false)
#define TRUE_IF_EVEN(x)     ((x & 1) ? false : true)

    REGISTER int val;

    val = __intVal(__INST(asciivalue));

    switch (val >> 8) {
        case 0x00:
            if ((unsigned)(val - 'A') <= ('Z' - 'A')) {
                RETURN ( true );
            }
            /* iso8859 puts national upper case characters at c0 .. df */
            if ((val >= 0xC0) && (val <= 0xDE)) {
                if (val != 0xD7) {
                    RETURN(true);
                }
            }
            RETURN (false);

        case 0x01:
            if (val <= 0x0137) { RETURN (TRUE_IF_EVEN(val)); }
            if (val <= 0x0148) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x0178) { RETURN (TRUE_IF_EVEN(val)); }
            if (val <= 0x017E) { RETURN (TRUE_IF_ODD(val)); } 
            if (val < 0x01CD) { 
                if (val == 0x0180) { RETURN (false); }
                if (val == 0x0181) { RETURN (true); }
                if (val <= 0x0186) {
                    RETURN (TRUE_IF_EVEN(val));
                }
                if (val <= 0x0189) {
                    RETURN (TRUE_IF_ODD(val));
                }
                if (val <= 0x018B) { RETURN (true); }
                if (val <= 0x018D) { RETURN (false); }
                if (val <= 0x0191) { RETURN (true); }
                if (val == 0x0193) { RETURN (true); }
                if (val == 0x0194) { RETURN (true); }
                if (val == 0x0196) { RETURN (true); }
                if (val == 0x0197) { RETURN (true); }
                if (val == 0x0198) { RETURN (true); }
                if (val == 0x019C) { RETURN (true); }
                if (val == 0x019D) { RETURN (true); }
                if (val == 0x019F) { RETURN (true); }
                if (val < 0x01A0) { RETURN (false); }
                if (val <= 0x01A6) { RETURN (TRUE_IF_EVEN(val)); }
                if (val <= 0x01AA) { RETURN (TRUE_IF_ODD(val)); }
                if (val <= 0x01AE) { RETURN (TRUE_IF_EVEN(val)); }
                if (val == 0x01B2) { RETURN (true); }
                if (val <= 0x01B7) { RETURN (TRUE_IF_ODD(val)); }
                if (val == 0x01B8) { RETURN (true); }
                if (val == 0x01BC) { RETURN (true); }
                if (val == 0x01C4) { RETURN (true); }
                if (val == 0x01C7) { RETURN (true); }
#if 0
                if (val == 0x01C8) { RETURN (true); }
#endif
                if (val == 0x01CA) { RETURN (true); }
#if 0
                if (val == 0x01CB) { RETURN (true); }
#endif
                RETURN (false);         /* WRONG !!! */ 
            }
            if (val <= 0x01DC) { RETURN (TRUE_IF_ODD(val)); } 
            if (val <= 0x01EF) { RETURN (TRUE_IF_EVEN(val)); }
            if (val == 0x01F0) { RETURN (false); }
            if (val == 0x01F1) { RETURN (true); }
            if (val == 0x01F2) { RETURN (false); }
            if (val == 0x01F3) { RETURN (false); }
            if (val == 0x01F4) { RETURN (true); }
            if (val == 0x01F5) { RETURN (false); }
            if (val == 0x01F6) { RETURN (true); }
            if (val == 0x01F7) { RETURN (true); } 
            RETURN (TRUE_IF_EVEN(val));

        case 0x02:
            if (val <= 0x0233) { RETURN (TRUE_IF_EVEN(val)); }
            RETURN (false); 

        case 0x03:
            if (val < 0x0386) { RETURN (false); }
            if (val == 0x0387) { RETURN (false); }
            if (val == 0x0390) { RETURN (false); }
            if (val <= 0x03AB) { RETURN (true); }
            if (val <= 0x03D1) { RETURN (false); }
            if (val <= 0x03D4) { RETURN (true); }
            if (val <= 0x03D7) { RETURN (false); }
            if (val <= 0x03EF) { RETURN (TRUE_IF_EVEN(val)); }
            if (val == 0x03F4) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val == 0x03F7) { RETURN (true); }
            if (val == 0x03F9) { RETURN (true); }
            if (val == 0x03Fa) { RETURN (true); }
#endif
            RETURN (false); 

        case 0x04:
            if (val <= 0x042F) { RETURN (true); }
            if (val <= 0x045F) { RETURN (false); }
            if (val <= 0x0481) { RETURN (TRUE_IF_EVEN(val)); }
            if (val < 0x048A) { RETURN (false); }
            if (val <= 0x04C0) { RETURN (TRUE_IF_EVEN(val)); }
#ifdef UNICODE_3_2
            if (val == 0x04C5) { RETURN (false); }
#endif
#ifdef UNICODE_3_2
            if (val <= 0x04C8) { RETURN (TRUE_IF_ODD(val)); }
#else
            if (val <= 0x04CA) { RETURN (TRUE_IF_ODD(val)); }
            if (val == 0x04CD) { RETURN (true); }
            if (val == 0x04CE) { RETURN (false); }
#endif
            if (val == 0x04CB) { RETURN (true); }
            if (val == 0x04CC) { RETURN (false); }
            RETURN (TRUE_IF_EVEN(val));

        case 0x05:
            if (val <= 0x050F) { RETURN (TRUE_IF_EVEN(val)); }
            if (val < 0x0531) { RETURN (false); }
            if (val <= 0x0556) { RETURN (true); }
            RETURN (false);

        case 0x10:
            if (val < 0x10A0) { RETURN (false); }
            if (val <= 0x10CF) { RETURN (true); }
            RETURN (false);

        case 0x1E:
            if (val < 0x1E96) { RETURN (TRUE_IF_EVEN(val)); }
            if (val < 0x1EA0) { RETURN (false); }
            RETURN (TRUE_IF_EVEN(val)); 

        case 0x1F:
            if (val <= 0x1F6F) {
                if (val & 0x0008) { RETURN (true); }
            }
            if (val <= 0x1F87) { RETURN (false); }
            if (val < 0x1FB8) { RETURN (false); }
            if (val < 0x1FBC) { RETURN (true); }
            if (val == 0x1FEC) { RETURN (true); }
            if (((val & 0x000F) >= 0x0008) && ((val & 0x000F) <= 0x000B)) { RETURN (true); }
            RETURN (false);
    
        case 0x21:
            if (val == 0x2102) { RETURN (true); }
            if (val == 0x2107) { RETURN (true); }
            if (val < 0x210B) { RETURN (false); }
            if (val < 0x210E) { RETURN (true); }
            if (val == 0x2110) { RETURN (true); }
            if (val == 0x2111) { RETURN (true); }
            if (val == 0x2112) { RETURN (true); }
            if (val == 0x2115) { RETURN (true); }
            if (val == 0x2119) { RETURN (true); }
            if (val == 0x211A) { RETURN (true); }
            if (val == 0x211B) { RETURN (true); }
            if (val == 0x211C) { RETURN (true); }
            if (val == 0x211D) { RETURN (true); }
            if (val < 0x2124) { RETURN (false); }
            if (val <= 0x212A) { RETURN (TRUE_IF_EVEN(val)); }
            if (val == 0x212B) { RETURN (true); }
            if (val == 0x212C) { RETURN (true); }
            if (val == 0x212D) { RETURN (true); }
            if (val == 0x2130) { RETURN (true); }
            if (val == 0x2131) { RETURN (true); }
            if (val == 0x2133) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val == 0x213E) { RETURN (true); }
            if (val == 0x213F) { RETURN (true); }
            if (val == 0x2145) { RETURN (true); }
#endif
            RETURN (false); 

        case 0xFF:
            if ((val >= 0xFF21) && (val <= 0xFF3A)) { RETURN (true); }
            RETURN (false);

        case 0x104:
            if (val <= 0x10427) { RETURN (true); }
            RETURN (false);

        case 0x1D4:
        case 0x1D5:
        case 0x1D6:
            if (val <= 0x1D419) { RETURN (true); }
            if (val < 0x1D6be) {
                if (((val - 0x1D41A) % 52) <= 25) { RETURN (false); }
                RETURN (true);
            }
            if (val < 0x1D6c1) { RETURN (true); }
            if (val < 0x1D6e2) { RETURN (false); }
            if (val < 0x1D6fb) { RETURN (true); }

            RETURN (false);
    }
    RETURN (false);

#undef TRUE_IF_ODD     
#undef TRUE_IF_EVEN    
%}
!

isVowel
    "return true, if I am a vowel (lower- or uppercase)"

    "/ I know the code is ugly;
    "/ better code is:
    "/     'aeiou' includes:self asLowercase
    "/ or:
    "/     'aeiouAEIOU' includes:self
    "/
    "/ until I have a smart compiler, I use the shorter (codewise):

    (self == $a) ifTrue:[^ true].
    (self == $e) ifTrue:[^ true].
    (self == $i) ifTrue:[^ true].
    (self == $o) ifTrue:[^ true].
    (self == $u) ifTrue:[^ true].
    (self == $A) ifTrue:[^ true].
    (self == $E) ifTrue:[^ true].
    (self == $I) ifTrue:[^ true].
    (self == $O) ifTrue:[^ true].
    (self == $U) ifTrue:[^ true].
    ^ false
! !

!Character methodsFor:'testing - national'!

isNationalAlphaNumeric
    "return true, if the receiver is a letter or digit in the
     current language (Language variable)"

    self isNationalLetter ifTrue:[^ true].
    ^ self isNationalDigit 
!

isNationalDigit
    "return true, if the receiver is a digit.
     This assumes unicode encoding.
     WARNING: this method is not complete."

    |codePoint "{ Class SmallInteger }"|

    codePoint := asciivalue.

    codePoint <= 16rFF ifTrue:[                "/ u00xx - unicode latin1 page 
        (codePoint between:($0 codePoint) and:($9 codePoint)) ifTrue:[^ true].
        ^ false
    ].

    (codePoint between:16rFF10 and:16rFF19) ifTrue:[ ^ true].
    ^ false.
!

isNationalLetter
    "return true, if the receiver is a letter.
     CAVEAT:
        for now, this method is only correct for unicode characters up to u+1d6ff (Unicode3.1).
        (which is more than mozilla does, btw. ;-)"

%{  /* NOCONTEXT */

    REGISTER int val;

    val = __intVal(__INST(asciivalue));
    switch (val >> 8) {
        case 0x00:
            if ((unsigned)(val - 'A') <= ('Z' - 'A')) {
                RETURN ( true );
            }
            if ((unsigned)(val - 'a') <= ('z' - 'a')) {
                RETURN ( true );
            }
            if (val == 0xAA) { RETURN (true); }
            if (val == 0xB5) { RETURN (true); }
            if (val == 0xBA) { RETURN (true); }
            if (val < 0xC0) { RETURN (false); }
            if (val == 0xD7) { RETURN (false); }
            if (val == 0xF7) { RETURN (false); }
            RETURN (true);

        case 0x01:                            
            RETURN (true);

        case 0x02:                            
#ifdef UNICODE_3_2
            if (val <= 0x2B8) { RETURN (true); }
            if (val == 0x2B9) { RETURN (false); }
            if (val == 0x2BA) { RETURN (false); }
#else
            if (val <= 0x2BA) { RETURN (true); }
#endif
            if (val <= 0x2C1) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val <= 0x2C5) { RETURN (false); }
            if (val <= 0x2CF) { RETURN (true); }
#endif
            if (val == 0x2D0) { RETURN (true); }
            if (val == 0x2D1) { RETURN (true); }
            if (val <= 0x2DF) { RETURN (false); }
            if (val <= 0x2E4) { RETURN (true); }
            if (val == 0x2EE) { RETURN (true); }
            RETURN (false);

        case 0x03:                            
            if (val == 0x37A) { RETURN (true); }
            if (val <= 0x385) { RETURN (false); }
            if (val == 0x387) { RETURN (false); }
#ifndef UNICODE_3_2
            if (val == 0x3F6) { RETURN (false); }
#endif
            RETURN (true);

        case 0x04:                            
            if (val <= 0x481) { RETURN (true); }
            if (val <= 0x486) { RETURN (false); }
            if (val == 0x487) { RETURN (true); }
#ifdef UNICODE_3_2
            if (val <= 0x48A) { RETURN (false); }
#else
            if (val <= 0x489) { RETURN (false); }
#endif
            RETURN (true);

        case 0x05:                            
            if (val <= 0x50f) { RETURN (true); }
            if (val <= 0x530) { RETURN (false); }
            if (val <= 0x556) { RETURN (true); }
            if (val <= 0x558) { RETURN (false); }
            if (val <= 0x559) { RETURN (true); }
            if (val <= 0x55F) { RETURN (false); }
            if (val <= 0x587) { RETURN (true); }
            if (val <= 0x5cf) { RETURN (false); } 
            if (val <= 0x5f2) { RETURN (true); }
            RETURN (false);

        case 0x06:                            
            if (val <= 0x620) { RETURN (false); }
            if (val <= 0x64A) { RETURN (true); }
            if (val <= 0x66D) { RETURN (false); }
            if (val == 0x670) { RETURN (false); }
            if (val <= 0x6D3) { RETURN (true); }
            if (val == 0x6D5) { RETURN (true); }
            if (val == 0x6E5) { RETURN (true); }
            if (val == 0x6E6) { RETURN (true); }
            if (val == 0x6EE) { RETURN (true); }
            if (val == 0x6EF) { RETURN (true); }
            if (val == 0x6FA) { RETURN (true); }
            if (val == 0x6FB) { RETURN (true); }
            if (val == 0x6FC) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val == 0x6FF) { RETURN (true); }
#endif
            RETURN (false);

        case 0x07:                            
            if (val <= 0x70F) { RETURN (false); }
            if (val == 0x711) { RETURN (false); }
            if (val <= 0x72F) { RETURN (true); }
#ifdef UNICODE_3_2
            if (val <= 0x74d) { RETURN (false); }
            if (val <= 0x74e) { RETURN (true); }
#else
            if (val <= 0x74c) { RETURN (false); }
            if (val <= 0x74f) { RETURN (true); }
#endif
            if (val <= 0x77F) { RETURN (false); }
            if (val <= 0x7a5) { RETURN (true); }
            if (val <= 0x7af) { RETURN (false); }
#ifndef UNICODE_3_2
            if (val == 0x7B1) { RETURN (true); }
#endif
            RETURN (false);

        case 0x09:                            
#ifdef UNICODE_3_2
            if (val <= 0x904) { RETURN (false); }
#else
            if (val <= 0x903) { RETURN (false); }
#endif
            if (val <= 0x93B) { RETURN (true); }
            if (val == 0x93D) { RETURN (true); }
            if (val == 0x950) { RETURN (true); }
            if (val <= 0x957) { RETURN (false); }
            if (val <= 0x961) { RETURN (true); }
            if (val <= 0x984) { RETURN (false); }
            if (val <= 0x9BB) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val == 0x9BD) { RETURN (true); }
#endif
            if (val <= 0x9DB) { RETURN (false); }
            if (val <= 0x9E1) { RETURN (true); }
            if (val <= 0x9EF) { RETURN (false); }
            if (val <= 0x9F1) { RETURN (true); }
            RETURN (false);

        case 0x0A:                            
            if (val <= 0xa04) { RETURN (false); }
            if (val <= 0xa3B) { RETURN (true); }
            if (val <= 0xa58) { RETURN (false); }
            if (val <= 0xa65) { RETURN (true); }
            if (val <= 0xa71) { RETURN (false); }
            if (val <= 0xa80) { RETURN (true); }
            if (val <= 0xa84) { RETURN (false); }
            if (val <= 0xaBB) { RETURN (true); }
            if (val == 0xaBD) { RETURN (true); }
            if (val <= 0xaCF) { RETURN (false); }
#ifndef UNICODE_3_2
            if (val == 0xAE2) { RETURN (false); }
            if (val == 0xAE3) { RETURN (false); }
#endif
            if (val <= 0xaE5) { RETURN (true); }
            RETURN (false);

        case 0x0B:                            
            if (val <= 0xB04) { RETURN (false); }
            if (val <= 0xb3B) { RETURN (true); }
            if (val == 0xb3d) { RETURN (true); }
            if (val <= 0xb5B) { RETURN (false); }
            if (val <= 0xb65) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val == 0xB71) { RETURN (true); }
            if (val == 0xB83) { RETURN (true); }
#endif
            if (val <= 0xb84) { RETURN (false); }
            if (val <= 0xbBB) { RETURN (true); }
            RETURN (false);

        case 0x0c:                            
            if (val <= 0xc04) { RETURN (false); }
            if (val <= 0xc3d) { RETURN (true); }
            if (val <= 0xc5f) { RETURN (false); }
            if (val <= 0xc65) { RETURN (true); }
            if (val <= 0xc84) { RETURN (false); }
#ifndef UNICODE_3_2
            if (val == 0xcbc) { RETURN (false); }
#endif
            if (val <= 0xcbd) { RETURN (true); }
            if (val <= 0xcdc) { RETURN (false); }
            if (val <= 0xce5) { RETURN (true); }
            RETURN (false);

        case 0x0d:                            
            if (val <= 0xd04) { RETURN (false); }
            if (val <= 0xd3d) { RETURN (true); }
            if (val <= 0xd5f) { RETURN (false); }
            if (val <= 0xd65) { RETURN (true); }
            if (val <= 0xd84) { RETURN (false); }
            if (val <= 0xdc9) { RETURN (true); }
            RETURN (false);

        case 0x0E:                            
            if (val == 0xE31) { RETURN (false); }
            if (val <= 0xE33) { RETURN (true); }
            if (val <= 0xE3F) { RETURN (false); }
            if (val <= 0xE46) { RETURN (true); }
            if (val <= 0xE7f) { RETURN (false); }
            if (val <= 0xEb0) { RETURN (true); }
            if (val == 0xEb1) { RETURN (false); }
            if (val <= 0xEb3) { RETURN (true); }
            if (val <= 0xEbc) { RETURN (false); }
            if (val <= 0xEc7) { RETURN (true); }
            if (val <= 0xEdb) { RETURN (false); }
            RETURN (true); 

        case 0x0F:                            
            if (val == 0xf00) { RETURN (true); }
            if (val <= 0xf3F) { RETURN (false); }
            if (val <= 0xf70) { RETURN (true); }
            if (val <= 0xf87) { RETURN (false); }
            if (val <= 0xf8f) { RETURN (true); }
            RETURN (false);

        case 0x10:                            
            if (val <= 0x102b) { RETURN (true); }
            if (val <= 0x104f) { RETURN (false); }
            if (val <= 0x1055) { RETURN (true); }
            if (val <= 0x109f) { RETURN (false); }
            if (val <= 0x10fa) { RETURN (true); }
            RETURN (false);

        case 0x11:                            
        case 0x12:                            
            RETURN (true);

        case 0x13:
            if (val <= 0x1360) { RETURN (true); }
            if (val <= 0x139f) { RETURN (false); }
            RETURN (true);

        case 0x14:                            
        case 0x15:                            
            RETURN (true);

        case 0x16:                            
            if (val == 0x166d) { RETURN (false); }
            if (val == 0x166e) { RETURN (false); }
            if (val == 0x1680) { RETURN (false); }
            if (val == 0x169b) { RETURN (false); }
            if (val == 0x169c) { RETURN (false); }
            if (val <= 0x16ea) { RETURN (true); }
            RETURN (false);

        case 0x17:
#ifndef UNICODE_3_2
            if (val == 0x1712) { RETURN (false); }
            if (val == 0x1713) { RETURN (false); }
            if (val == 0x1714) { RETURN (false); }
            if (val == 0x1732) { RETURN (false); }
            if (val == 0x1733) { RETURN (false); }
            if (val == 0x1734) { RETURN (false); }
            if (val == 0x1735) { RETURN (false); }
            if (val == 0x1736) { RETURN (false); }
            if (val == 0x1752) { RETURN (false); }
            if (val == 0x1753) { RETURN (false); }
            if (val == 0x1772) { RETURN (false); }
            if (val == 0x1773) { RETURN (false); }
#endif
            if (val <= 0x17b3) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val == 0x17D7) { RETURN (true); }
            if (val == 0x17DC) { RETURN (true); }
#endif
            RETURN (false);

        case 0x18:
            if (val <= 0x181f) { RETURN (false); }
            if (val <= 0x18a8) { RETURN (true); }
            RETURN (false);

        case 0x19:
#ifndef UNICODE_3_2
            if (val <= 0x191F) { RETURN (true); }
            if (val <= 0x194F) { RETURN (false); }
            if (val <= 0x197F) { RETURN (true); }
#endif
            RETURN (false);

        case 0x1d:
            if (val = 0x1d00) { RETURN (true); }
            RETURN (false);

        case 0x1e:
            RETURN (true);

        case 0x1f:
            if (val <= 0x1fbc) { RETURN (true); }
            if (val == 0x1fbe) { RETURN (true); }
            if (val <= 0x1fc1) { RETURN (false); }
            if (val <= 0x1fcc) { RETURN (true); }
            if (val <= 0x1fcf) { RETURN (false); }
            if (val <= 0x1fdc) { RETURN (true); }
            if (val <= 0x1fdf) { RETURN (false); }
            if (val <= 0x1fec) { RETURN (true); }
            if (val <= 0x1ff1) { RETURN (false); }
            if (val <= 0x1ffc) { RETURN (true); }
            RETURN (false);

        case 0x20:
#ifndef UNICODE_3_2
            if (val == 0x2071) { RETURN (true); }
#endif
            if (val == 0x207f) { RETURN (true); }
            RETURN (false);

        case 0x21:
            if (val == 0x2102) { RETURN (true); }
            if (val == 0x2107) { RETURN (true); }
            if (val <= 0x2109) { RETURN (false); }
            if (val <= 0x2113) { RETURN (true); }
            if (val == 0x2115) { RETURN (true); }
            if (val <= 0x2118) { RETURN (false); }
            if (val <= 0x211d) { RETURN (true); }
            if (val <= 0x2123) { RETURN (false); }
            if (val == 0x2125) { RETURN (false); }
            if (val == 0x2127) { RETURN (false); }
            if (val == 0x2129) { RETURN (false); }
            if (val == 0x212E) { RETURN (false); }
            if (val == 0x2132) { RETURN (false); }
            if (val == 0x213A) { RETURN (false); }
#ifndef UNICODE_3_2
            if (val == 0x213B) { RETURN (false); }
            if (val <= 0x213F) { RETURN (true); }
            if (val <= 0x2144) { RETURN (false); }
            if (val == 0x214A) { RETURN (false); }
            if (val == 0x214B) { RETURN (false); }
#endif
            if (val <= 0x2152) { RETURN (true); }
            RETURN (false);

        case 0x30:
            if (val == 0x3005) { RETURN (true); }
            if (val == 0x3006) { RETURN (true); }
            if (val <= 0x3030) { RETURN (false); }
            if (val <= 0x3035) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val == 0x303B) { RETURN (true); }
            if (val == 0x303C) { RETURN (true); }
#endif
            if (val <= 0x3040) { RETURN (false); }
            if (val <= 0x3098) { RETURN (true); }
            if (val <= 0x309c) { RETURN (false); }
#ifndef UNICODE_3_2
            if (val == 0x30A0) { RETURN (false); }
#endif
            if (val == 0x30Fb) { RETURN (false); }
            RETURN ((true));

        case 0x31:
            if (val <= 0x318f) { RETURN (true); }
            if (val <= 0x319F) { RETURN (false); }
            RETURN ((true));

        case 0x34:
            RETURN ((true));

        case 0x4d:
            if (val <= 0x4DB4) { RETURN (false); }
#ifndef UNICODE_3_2
            if (val <= 0x4DBF) { RETURN (true); }
            RETURN (false);
#else
            RETURN (true);
#endif

        case 0x4e:
            RETURN ((true));

        case 0x9f:
            if (val <= 0x9fa4) { RETURN (false); }
            RETURN (true);

        case 0xA0:
        case 0xA1:
        case 0xA2:
        case 0xA3:
            RETURN (true);

        case 0xA4:
            if (val <= 0xa48f) { RETURN (true); }
            RETURN (false);

        case 0xA5:
            RETURN (true);

        case 0xAC:
            RETURN (true);

        case 0xD7:
            RETURN (true);

        case 0xF9:
        case 0xFA:
            RETURN (true);

        case 0xFB:
            if (val == 0xfb1e) { RETURN (false); }
            if (val == 0xfb29) { RETURN (false); }
            RETURN (true);

        case 0xFC:
            RETURN (true);

        case 0xFD:
            if (val <= 0xFD3d) { RETURN (true); }
            if (val <= 0xFD4F) { RETURN (false); }
#ifndef UNICODE_3_2
            if (val == 0xFDFC) { RETURN (false); }
            if (val == 0xFDFD) { RETURN (false); }
#endif
            RETURN (true);

        case 0xFE:
#ifndef UNICODE_3_2
            if (val <= 0xFE0F) { RETURN (false); }
#endif
            if (val <= 0xFE1f) { RETURN (true); }
            if (val <= 0xFE6F) { RETURN (false); }
            if (val <= 0xFEFE) { RETURN (true); }
            RETURN (false);

        case 0xFF:                            
            if (val <= 0xFF20) { RETURN (false); }
            if (val <= 0xFF3a) { RETURN (true); }
            if (val <= 0xFF40) { RETURN (false); }
            if (val <= 0xFF5a) { RETURN (true); }
            if (val <= 0xFF65) { RETURN (false); }
            if (val <= 0xFFdC) { RETURN (true); }
            RETURN (false);

        case 0x100:                            
#ifndef UNICODE_3_2
            RETURN (true);
#else
            RETURN (false);
#endif

        case 0x103:                            
            if (val <= 0x1031f) { RETURN (true); }
            if (val <= 0x1032F) { RETURN (false); }
            if (val <= 0x10349) { RETURN (true); }
#ifndef UNICODE_3_2
            if (val <= 0x1037F) { RETURN (false); }
            if (val <= 0x1039E) { RETURN (true); }
#endif
            RETURN (false);

        case 0x104:                            
#ifndef UNICODE_3_2
            if (val <= 0x1049F) { RETURN (true); }
            if (val <= 0x104aF) { RETURN (false); }
#endif
            RETURN (true);

        case 0x108:                            
#ifndef UNICODE_3_2
            RETURN (true); 
#else
            RETURN (false);
#endif

        case 0x1D4:                            
        case 0x1D5:                            
            RETURN (true);

        case 0x1D6:                            
            if (val == 0x1d6c1) { RETURN (false); }
            if (val == 0x1d6db) { RETURN (false); }
            if (val == 0x1d6fb) { RETURN (false); }
            RETURN (true);
    }
    RETURN (false);
%}
! !

!Character methodsFor:'tracing'!

traceInto:aRequestor level:level from:referrer
    "double dispatch into tracer, passing my type implicitely in the selector"

    ^ aRequestor traceCharacter:self level:level from:referrer

! !

!Character methodsFor:'visiting'!

acceptVisitor:aVisitor with:aParameter

    ^ aVisitor visitCharacter:self with:aParameter
! !

!Character class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic/Character.st,v 1.123 2005-04-08 22:49:09 stefan Exp $'
! !
